#include "processing.h"
#include "utils.h"

#include "runtime/file_tab.h"
#include "runtime/threading.h"
#include "runtime/aio.h"
#include "runtime/buffer_manager.h"

#include <string.h>
#include <assert.h>

typedef struct {
    vy_file_id fid;

    /* How many times has this file been added? */
    unsigned int turn;
} vy_file_processing_queue_entry;

#define QUEUE_LENGTH 1024
typedef struct {
    vy_file_processing_queue_entry entries[QUEUE_LENGTH];
    unsigned int head;
    unsigned int tail;
} vy_file_processing_queue;

static vy_file_processing_queue _processing_queue;
static vy_mutex *_guard;
static bool _keep_running;


static vy_result vyAddFileToProcessingQueueImpl(vy_file_id file, unsigned int turn) {
    if (!_guard)
        _guard = vyCreateMutex();

    vy_result result = VY_SUCCESS;

    vy_file_processing_queue_entry entry = {
        .fid  = file,
        .turn = turn,
    };
    vyLockMutex(_guard);
    if ((_processing_queue.tail + 1) % QUEUE_LENGTH != _processing_queue.head) {
        unsigned int slot               = _processing_queue.head;
        _processing_queue.entries[slot] = entry;
        _processing_queue.head = (_processing_queue.head + 1) % QUEUE_LENGTH;
    } else {
        vyReportError("ASSETC", "The processing queue is full!");
        result = 1;
    }
    vyUnlockMutex(_guard);
    return result;
}


vy_result vyAddFileToProcessingQueue(vy_file_id file) {
    return vyAddFileToProcessingQueueImpl(file, 1);
}


#define MAX_PROCESSORS 256
static vy_processor_fn *_processor_fns[MAX_PROCESSORS];
static const char *_processor_exts[MAX_PROCESSORS];
static unsigned int _processor_count;

vy_result vyAddAssetProcessor(const char *file_extension, vy_processor_fn fn) {
    /* Should only be called from main thread */
    if (_processor_count == MAX_PROCESSORS) {
        vyReportError("ASSETC", "Too many asset processor functions!");
        return 1;
    }
    _processor_fns[_processor_count] = fn;
    _processor_exts[_processor_count] = file_extension;
    ++_processor_count;
    return VY_SUCCESS;
}

static void PopAndSwapSubmittedData(unsigned int at,
                                    unsigned int *count,
                                    vy_file_processing_queue_entry *queue_entries,
                                    vy_aio_handle *handles,
                                    void **buffers,
                                    size_t *sizes) {
    if (at < *count - 1) {
        queue_entries[at] = queue_entries[*count - 1];
        buffers[at]       = buffers[*count - 1];
        handles[at]       = handles[*count - 1];
        sizes[at]         = sizes[*count - 1];
    }
    *count = *count - 1;
}

static void ProcessLoadedFile(vy_file_processing_queue_entry entry, void *buffer, size_t size) {
    /* Search for a matching processor function */
    const char *path = vyGetFilePath(entry.fid);
    size_t path_len = strlen(path);
    for (unsigned int i = 0; i < _processor_count; ++i) {
        size_t ext_len = strlen(_processor_exts[i]);
        if (ext_len > path_len)
            continue;

        const char *path_end = &path[path_len - ext_len];
        if (memcmp(path_end, _processor_exts[i], ext_len) == 0) {
            vy_processor_output out;
            vy_result res = _processor_fns[i](entry.fid, buffer, size, &out);
            if (res == VY_PROCESSING_FAILED) {
                vyLog("ASSETC", "Failed to process file: %s", path);
            } else if (res == VY_PROCESSING_TRY_AGAIN) {
                if (entry.turn < 2) {
                    vyAddFileToProcessingQueueImpl(entry.fid, entry.turn + 1);
                } else {
                    vyLog("ASSETC",
                          "File '%s' took too many turns to process: %u",
                          path,
                          entry.turn);
                }
            }
            return;
        }
    }
    vyLog("ASSETC", "No asset processor for file: %s", path);
}

static void ProcessingThread(void *_param) {
    VY_UNUSED(_param);

    
    vy_file_processing_queue_entry submitted_entries[VY_LOAD_BATCH_MAX_SIZE];
    vy_aio_handle submitted_handles[VY_LOAD_BATCH_MAX_SIZE];
    void *submitted_buffers[VY_LOAD_BATCH_MAX_SIZE];
    size_t submitted_sizes[VY_LOAD_BATCH_MAX_SIZE];
    unsigned int submitted_outstanding = 0;

    while (_keep_running) {
        vy_load_batch load_batch;
        vy_file_processing_queue_entry load_entries[VY_LOAD_BATCH_MAX_SIZE];
        void *load_buffers[VY_LOAD_BATCH_MAX_SIZE];
        size_t load_sizes[VY_LOAD_BATCH_MAX_SIZE];
        load_batch.num_loads = 0;

        bool got_entry = false;
        do {
            got_entry = false;
            vy_file_processing_queue_entry entry;
            vyLockMutex(_guard);
            if (_processing_queue.head != _processing_queue.tail) {
                unsigned int next = _processing_queue.tail;
                entry             = _processing_queue.entries[next];
                _processing_queue.tail =
                    (_processing_queue.tail + 1) % QUEUE_LENGTH;
                got_entry = true;
            }
            vyUnlockMutex(_guard);
            if (!got_entry)
                continue;

            const char *path = vyGetFilePath(entry.fid);
            if (!path) {
                vyLog("ASSETC", "Invalid file id: %#x", entry.fid);
                continue;
            }

            size_t fsz = vyGetFileSize(path);
            void *dest = vyAllocBuffer(fsz);
            if (!dest) {
                vyLog("ASSETC",
                      "Ran out of memory for loading the file: %s",
                      path);
                continue;
            }

            load_sizes[load_batch.num_loads]                 = fsz;
            load_buffers[load_batch.num_loads]               = dest;
            load_batch.loads[load_batch.num_loads].file      = entry.fid;
            load_batch.loads[load_batch.num_loads].dest      = dest;
            load_batch.loads[load_batch.num_loads].num_bytes = fsz;
            load_batch.loads[load_batch.num_loads].offset    = 0;
            load_entries[load_batch.num_loads]               = entry;
            ++load_batch.num_loads;
        }
        while (got_entry && load_batch.num_loads < VY_LOAD_BATCH_MAX_SIZE);

        vy_aio_handle load_handles[VY_LOAD_BATCH_MAX_SIZE];
        if (load_batch.num_loads > 0) {
            vy_result submit_result =
                vySubmitLoadBatch(&load_batch, load_handles);
            if (submit_result != VY_SUCCESS) {
                vyLog("ASSETC", "SubmitLoadBatch failed: %u", submit_result);
                continue;
            }
        }


        /* Process the previously submitted loads */
        while (submitted_outstanding > 0) {
            for (unsigned int i = 0; i < submitted_outstanding; ++i) {
                vy_aio_state state = vyGetAIOState(submitted_handles[i]);
                switch (state) {
                case VY_AIO_STATE_PENDING:
                    continue;
                case VY_AIO_STATE_FAILED:
                    vyLog("ASSETC",
                          "Loading file %s failed.",
                          vyGetFilePath(submitted_entries[i].fid));
                    vyReleaseBuffer(submitted_buffers[i], submitted_sizes[i]);
                    PopAndSwapSubmittedData(i,
                                            &submitted_outstanding,
                                            submitted_entries,
                                            submitted_handles,
                                            submitted_buffers,
                                            submitted_sizes);
                    --i;
                    break;
                case VY_AIO_STATE_INVALID:
                    vyLog("ASSETC",
                          "Got invalid AIO handle for file: %s",
                          vyGetFilePath(submitted_entries[i].fid));
                    vyReleaseBuffer(submitted_buffers[i], submitted_sizes[i]);
                    PopAndSwapSubmittedData(i,
                                            &submitted_outstanding,
                                            submitted_entries,
                                            submitted_handles,
                                            submitted_buffers,
                                            submitted_sizes);
                    --i;
                    break;
                case VY_AIO_STATE_FINISHED:
                    ProcessLoadedFile(submitted_entries[i],
                                      submitted_buffers[i],
                                      submitted_sizes[i]);
                    vyReleaseBuffer(submitted_buffers[i], submitted_sizes[i]);
                    PopAndSwapSubmittedData(i,
                                            &submitted_outstanding,
                                            submitted_entries,
                                            submitted_handles,
                                            submitted_buffers,
                                            submitted_sizes);
                    --i;
                }
            }
        }

        /* Start new round */
        assert(sizeof(submitted_entries) == sizeof(load_entries));
        assert(sizeof(submitted_handles) == sizeof(load_handles));
        assert(sizeof(submitted_buffers) == sizeof(load_buffers));
        assert(sizeof(submitted_sizes) == sizeof(load_sizes));
        memcpy(submitted_entries, load_entries, sizeof(submitted_entries));
        memcpy(submitted_handles, load_handles, sizeof(submitted_handles));
        memcpy(submitted_buffers, load_buffers, sizeof(submitted_buffers));
        memcpy(submitted_sizes, load_sizes, sizeof(submitted_sizes));
        submitted_outstanding = load_batch.num_loads;
    }
}

#define NUM_PROCESSING_THREADS 8

vy_thread *_processing_threads[NUM_PROCESSING_THREADS];

vy_result vyStartProcessing(void) {
    if (!_guard)
        _guard = vyCreateMutex();

    _keep_running = true;
    for (unsigned int i = 0; i < NUM_PROCESSING_THREADS; ++i) {
        _processing_threads[i] = vySpawnThread(ProcessingThread, NULL);
        if (!_processing_threads[i]) {
            vyReportError("ASSETC", "Failed to spawn processing thread %u!", i);
            _keep_running = false;
            return 1;
        }
    }
    return VY_SUCCESS;
}

void vyStopProcessing(void) {
    _keep_running = false;
    for (unsigned int i = 0; i < NUM_PROCESSING_THREADS; ++i)
        vyJoinThread(_processing_threads[i]);
}